
pragma SPARK_Mode (On);

with Types, Failure, Inertia, Clock, Watchdog, Warhead, Steer, Dummy;
use  Types;

procedure Main
  with
  
  -- global state modified by this procedure
  
  Global  => 
  
  (Input  => 
     (Failure.Failure_Detector, 
      Inertia.X_Acc_Meter, 
      Inertia.Y_Acc_Meter, 
      Inertia.Z_Ang_Meter, 
      Clock.Clock
     ),
   
   Output => 
     (Watchdog.Lifesign, 
      Warhead.Detonator, 
      Steer.Fin,
      Dummy.Dummy              -- added to circumvent relevance check on Failure.Failure_Detector.
     ) 
  ),
    
  -- global state dependencies the procedure creates
    
  Depends => 
    
  (Watchdog.Lifesign => 
     null,                     -- NOTHING affects self-destruction
   
   Warhead.Detonator => 
     (Inertia.X_Acc_Meter,     -- for checking if current location close enough to target
      Inertia.Y_Acc_Meter,     -- for checking if current location close enough to target
      Clock.Clock              -- for checking if current location close enough to target (integration)
     ),
   
   Steer.Fin =>
     (Inertia.X_Acc_Meter,     -- for pointing missile towards target relative to current location
      Inertia.Y_Acc_Meter,     -- for pointing missile towards target relative to current location
      Inertia.Z_Ang_Meter,     -- for pointing missile towards target relative to current location (compute orientation)
      Clock.Clock              -- for pointing missile towards target relative to current location (integration)
     ),
   
   Dummy.Dummy =>
     Failure.Failure_Detector  -- makes reads from Failure.Failure_Detector relevant.
  )
    
is
   
   -- temporary variables used by the procedure:
   
   -- sensor readings
   F : Boolean       := False;
   T : Time_Delta    := 0.0;
   X_Acc : Magnitude := 0.0;
   Y_Acc : Magnitude := 0.0;
   Z_Ang : Angle     := 0.0;

   -- helper variables
   Displacement     : Vector;
   Displacement_Ang : Angle;
   To_Target        : Vector;
   To_Target_Ang    : Angle;
   
   -- persistent variables used by the procedure:
   
   Location : Vector   := (X => 0.0, Y => 0.0);
   Destruct : Boolean  := False;
   Detonate : Boolean  := False;
   Orientation : Angle := Orientation_Initial;
   
begin
   
   -- sense-control-actuate loop
   
   loop
      
      -- sense
      Failure . Sense (F);
      Inertia . Sense (X_Acc, Y_Acc, Z_Ang);
      Clock   . Sense (T);
      
      -- control
      Displacement     := ( X => Integral ( Integral ( X_Acc , T ) , T ) , 
                            Y => Integral ( Integral ( Y_Acc , T ) , T ) );
      Location         := Vector_Add (Location, Displacement);
      To_Target        := Vector_Sub ( Target , Location );
      Displacement_Ang := Integral ( Integral ( Z_Ang , T ) , T );
      Destruct         := Destruct or F;
      Detonate         := Detonate or (Vector_Mag (To_Target) <= Threshold);
      Orientation      := Angle_Normalize ( Orientation + Displacement_Ang );
      To_Target_Ang    := Vector_Ang (To_Target);
      
      -- actuate dummy
      Dummy   . Actuate ( Destruct );
      
      -- actuate steering
      Steer   . Actuate ( Angle_Normalize ( To_Target_Ang - Orientation ) );
      
      -- actuate watchdog
      if Destruct then
         -- stop progress completely, 
         -- thus stopping watchdog actuation, 
         -- thus forcing watchdog to self-destruct the missile.
         loop 
            null;
         end loop;
      end if;
      Watchdog.Actuate;
      
      -- actuate warhead
      if Detonate then
         -- detonate the nuclear warhead (instantly).
         Warhead.Actuate;
      end if;
      
      -- reset helper variables
      Displacement     := (X => 0.0, Y => 0.0);
      Displacement_Ang := 0.0;
      To_Target        := (X => 0.0, Y => 0.0);
      To_Target_Ang    := 0.0;
      
   end loop;
   
end Main;
