with Ada.Numerics.Generic_Elementary_Functions, Ada.Numerics;

package Types is
   
   -- elementary functions
   package EF is new Ada.Numerics.Generic_Elementary_Functions(Float);
   use EF;

   ------------------------------------------------------------------------
   -- Types and (stateless) operations on these types.
   ------------------------------------------------------------------------
   
   -- TYPES
   
   -- magnitude
   -- e.g. distance (m (meters)), speed (m/s), quickening (m/s^2).
   subtype Magnitude is Float;
   
   --  we would have liked to have different types for all these
   --  different magnitudes, but since there is no convenient type
   --  polymorphism in Ada, defining the different vectors
   --  (displacement, velocity, acceleration) and operations on them
   --  would have been a pain.
   
   -- angle
   -- e.g. direction (in radians), angular acceleration (in radians/s^2).
   subtype Angle is Float;
   
   -- time
   -- in seconds (including fractions).
   subtype Time_Delta is Float;
   
   -- vector (Cartesian)
   -- e.g. displacement, velocity, acceleration.
   type Vector is
      record
         X : Magnitude;
         Y : Magnitude;
      end record;
   
   ------------------------------------------------------------------------
   
   -- CONSTANTS
   
   -- pre-launch programming
   Target              : constant Vector    := (X => 0.0, Y => 1_000_000.0);
   Threshold           : constant Magnitude := 1_000.0;
   Orientation_Initial : constant Angle := 0.5 * Ada.Numerics.Pi;
   
   -- helper constants used thoughout code.
   Two_Pi              : constant Float := (2.0 * Ada.Numerics.Pi);
   
   ------------------------------------------------------------------------
   
   -- OPERATIONS
   
   -- computes the area of the Y by Xd rectangle
   -- (this rectangular area approximates the area under a quickening/speed 
   --  function between X (time last measured) and X+Xd, á la Riemann integral).
   function Integral ( Y  : Float ;  -- new Y
                       Xd : Float )  -- delta X
                     return Float
   is ( Y * Xd );
   
   function Vector_Add ( A  : Vector ;
                         B  : Vector)
                       return Vector
   is ( X => (A.X + B.X), Y => (A.Y + B.Y) );
   
   function Vector_Sub ( A  : Vector ;
                         B  : Vector)
                       return Vector
   is ( X => (A.X - B.X), Y => (A.Y - B.Y) );
   
   function Vector_Mag ( A  : Vector )
                       return Magnitude
   is ( Sqrt ( (A.X * A.X) + (A.Y * A.Y) ) );
   
   function Vector_Ang ( A  : Vector )
                       return Angle
   is ( 2.0 * ( Arctan ( A.Y / (A.X + Sqrt ( A.X * A.X + A.Y * A.Y) ) ) ) );
   
   -- no-op when angle is between 0 and 2*Pi.
   function Angle_Modulus ( A : Angle )
                            return Angle
   is ( 
     A - ( Two_Pi ** (Float'Floor ( Log (A, 2.0 * Ada.Numerics.Pi) )))
   );
   
   -- no-op when angle is positive.
   function Angle_Positive ( A : Angle )
                           return Angle
   is (
     A + ( Two_Pi ** (Float'Ceiling ( Log (A, 2.0 * Ada.Numerics.Pi) )))
   );
     
   function Angle_Normalize ( A : Angle )
                            return Angle
   is (
       Angle_Modulus ( Angle_Positive (A) )
   );
   
end Types;
